/***********************************************************
 * $Id: CompressionHelper.java 401 2011-01-01 09:29:20Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 17.09.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http;

import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.GZIPOutputStream;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class provides static functionality for handling compression algorithms
 * controlled through the <code>Content-Coding</code> header field. 
 */
public abstract class CompressionHelper
{
    private static final Logger log = LoggerFactory.getLogger(CompressionHelper.class);
    
    /**
     * A default set of supported encodings.
     */
    public static final String[] DEFAULT_SUPPORTED_ENCODINGS = new String[] {"gzip","deflate"};
    
    /**
     * 
     * @param acceptEncoding The content of an Accept-Encoding header (see section 14.3 of RFC1616).
     * @param supportedEncodings A list of supported encodings to choose from. This list should
     *                           not contain the <code>identity</code> encoding, because
     *                           noop compression is selected through a <code>null</code>
     *                           encoding value.
     * @return The encoding selected to use for the given browser of <code>null</code>, if the response
     *         should not be compressed.
     */
    public static String selectCompressionCoding(String acceptEncoding, String[] supportedEncodings) {
        
        String selectedEncoding = null;     
        
        if (acceptEncoding != null && supportedEncodings != null && supportedEncodings.length > 0) {
            
            String[] encodingsWithQ = acceptEncoding.split(",");
            double qSelected = -1.0;
            
            for (int i=0;i<encodingsWithQ.length;++i) {
                
                try {
                    String[] encq = encodingsWithQ[i].split(";",2);
                    
                    double q;
                    
                    if (encq.length > 1) {
                        
                        encq[1] = encq[1].trim();
                        if (encq[1].startsWith("q=")) 
                            q = Double.parseDouble(encq[1].substring(2));
                        else
                            throw new IllegalArgumentException("No valid q-value after semicolon.");
                    }
                    else
                        q = 1.0;
                    
                    if (q > qSelected) {
                        
                        encq[0] = encq[0].trim();
                
                        if ("*".equals(encq[0])) {
                            selectedEncoding = supportedEncodings[0];
                            qSelected = q;
                        }
                        else {
                            
                            for (String enc : supportedEncodings) {
                                if (encq[0].equals(enc)) {
                                    selectedEncoding = enc;
                                    qSelected = q;
                                    break;
                                }
                            }
                        }
                    }
                    
                } catch (Throwable e) {
                    log.error("Received invalid accept-encoding part ["+ encodingsWithQ[i]+"]: "+e.getMessage());
                }
            }
            if (qSelected <= 0.0)
                selectedEncoding = null;
        }
        return selectedEncoding;
    }
    
    /**
     * Parse the <code>Accept-Encoding</code> header of the request and set up a possibly compressed
     * output stream to write to. If the output is chosen to be compressed, an appropriate
     * <code>Content-Encoding</code> is set on the passed HTTP response.
     * 
     * @param req The HTTP request to examine.
     * @param resp The HTTP response.
     * @return A compressed or uncompressed stream based on the <code>Accept-Encoding</code>
     *         header of the client. The retuned stream should be finished
     *         after writing the whole response using {@link #finishOutputStream(OutputStream)}.
     * @throws IOException Upon errors when setting response headers or setting up streams.
     */
    public static OutputStream setupResponseOutputStream(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        
        String acceptEncoding = req.getHeader("Accept-Encoding");
        String selectedEncoding = selectCompressionCoding(acceptEncoding,DEFAULT_SUPPORTED_ENCODINGS); 
            
        if (selectedEncoding == null)
            return resp.getOutputStream();
        else {
            
            resp.setHeader("Content-Encoding",selectedEncoding);
            
            if ("deflate".equals(selectedEncoding)) {
        
                return new DeflaterOutputStream(resp.getOutputStream());
            }
            else {
                return new GZIPOutputStream(resp.getOutputStream());
            }
        }
    }
    
    public static boolean isCompressed(OutputStream os) {
        
        return os instanceof DeflaterOutputStream;
    }
    
    /**
     * Finish writing to a possibly compressed servlet output stream.
     * 
     * @param os An OutputStream returned by {@link #setupResponseOutputStream(HttpServletRequest, HttpServletResponse)}.
     * @throws IOException Upon I/O errors.
     */
    public static void finishOutputStream(OutputStream os) throws IOException {
        
        if (os instanceof DeflaterOutputStream) {
            DeflaterOutputStream dos = (DeflaterOutputStream) os;
            dos.finish();
        }
        else {
            os.flush();
        }
    }
  
    
}
