/***********************************************************
 * $Id: ResourceServlet.java 401 2011-01-01 09:29:20Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 17.09.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.codec.binary.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A servlet, which exports parts of the {@link ServletContext}'s resources as HTTP resources
 * with configuration options on which items should be delivered in compressed form.
 * 
 * <p>This servlet may be configured through initialization parameters inside a classical
 * servlet container like in the following <code>web.xml</code> example:
 * </p>
 * <pre>
 * &lt;servlet&gt;
 *  &lt;servlet-name&gt;foobar-resources&lt;/servlet-name&gt;
 *  &lt;servlet-class&gt;org.clazzes.util.http.ResourceServlet&lt;/servlet-class&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;resourcePath&lt;/param-name&gt;
 *   &lt;param-value&gt;/foobar&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;excludeMimeType&lt;/param-name&gt;
 *   &lt;param-value&gt;image/png,image/jpeg,image/gif&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;aliases&lt;/param-name&gt;
 *   &lt;param-value&gt;/=/index.html,/index.htm=/index.html&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 *  &lt;init-param&gt;
 *   &lt;param-name&gt;redirects&lt;/param-name&gt;
 *   &lt;param-value&gt;/app1=appServlet,/app2=/ctx2/appServlet&lt;/param-value&gt;
 *  &lt;/init-param&gt;
 * &lt;/servlet&gt;
 * </pre>
 * 
 * <p>Additional, seldom used initialization parameters are <code>minCompressionSize</code> and
 * <code>bufferSize</code>, which are actually integer values.</p>
 * 
 * <p>In OSGi frameworks it is recommended to set up this servlet using blueprint or a similar
 * technology using the supplied bean property setters like in the following example:</p>
 * <pre>
 *  &lt;bean id="resourceServlet" class="org.clazzes.util.http.ResourceServlet"&gt;
 *    &lt;property name="resourcePath" value="/foobar-resources"&gt;&lt;/property&gt;
 *    &lt;property name="excludeMimeTypes"&gt;
 *      &lt;set&gt;
 *        &lt;value&gt;image/png&lt;/value&gt;
 *        &lt;value&gt;image/gif&lt;/value&gt;
 *        &lt;value&gt;image/jpeg&lt;/value&gt;
 *        &lt;value&gt;image/tiff&lt;/value&gt;
 *      &lt;/set&gt;
 *    &lt;/property&gt;
 *    &lt;property name="aliases"&gt;
 *      &lt;map&gt;
 *        &lt;entry key="/"          value="/index.html"/&gt;       
 *        &lt;entry key="/index.htm" value="/index.html"/&gt;       
 *      &lt;/map&gt;
 *    &lt;/property&gt;
 *    &lt;property name="redirects"&gt;
 *      &lt;map&gt;
 *        &lt;entry key="/app1" value="appServlet"/&gt;       
 *        &lt;entry key="/app2" value="/ctx2/appServlet"/&gt;       
 *      &lt;/map&gt;
 *    &lt;/property&gt;
 *  &lt;/bean&gt;
 * </pre>
 */
public class ResourceServlet extends HttpServlet
{
    private static final Logger log = LoggerFactory.getLogger(ResourceServlet.class);
    
    private static final long serialVersionUID = 2854439871115810811L;

    private String lastModified;
    private int bufferSize = 4096;
    private String resourcePath;
    private int minCompressionSize = 4096;
    private Set<String> excludeMimeTypes;
    private Map<String,String> redirects;
    private Map<String,String> aliases;
   
    private ConcurrentMap<String,ResourceInfo> resourceInfos;
    
    private static final class ResourceInfo {
        
        public ResourceInfo(String eTag, long contentLength)
        {
            super();
            this.eTag = eTag;
            this.contentLength = contentLength;
        }
        
        public final String eTag;
        public final long contentLength;
        
        public String toString() {
            return "eTag="+this.eTag+",contentLength="+this.contentLength;
        }
    }
    
    public ResourceServlet() {
        this.bufferSize = 4096;
        this.resourceInfos = new ConcurrentHashMap<String,ResourceInfo>();
        
        SimpleDateFormat df = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z",Locale.US);
        
        this.lastModified = df.format(new Date());
        
        if (log.isDebugEnabled())
            log.debug("Servlet created, lastModified=["+this.lastModified+"].");
    }
    
    
    
    private ResourceInfo enterETag(String pathInfo, InputStream is, byte[] buf) throws IOException, ServletException {
        
        int n;
        try
        {
            MessageDigest md = MessageDigest.getInstance("SHA-1");
            long l = 0;
            
            while ((n=is.read(buf)) > 0) {
            
                md.update(buf,0,n);
                l+=n;
            }
            
            Base64 b64 = new Base64(80,new byte[0]);
            String eTag = b64.encodeToString(md.digest());
            
            ResourceInfo ri = new ResourceInfo(eTag,l);
            this.resourceInfos.putIfAbsent(pathInfo,ri);
            return ri;

        } catch (NoSuchAlgorithmException e)
        {
            throw new ServletException("SHA-1 digest not found",e);
        }
       
    }
    
    /* (non-Javadoc)
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException,
                                                                          IOException
    {
        String path = req.getPathInfo();
        
        // some servlet container send a null path info, if the servlet path
        // itself is specified without any extra path and without a slash.
        // replace null with empty string in order to allow easier configuration
        // of redirects and aliases for the root context.
        if (path == null) path="";
        
        if (log.isDebugEnabled())
            log.debug("Received request ["+path+"]");

        if (this.redirects != null) {
            
            String redirect = this.redirects.get(path);
            
            if (redirect != null) {
             
                if (log.isDebugEnabled())
                    log.debug("Redirecting request ["+path+"] to ["+redirect+"].");
                
                resp.sendRedirect(redirect);
                return;
            }
        }
        
        InputStream is = null;
        OutputStream os = null;
        
        if (this.resourcePath == null)
            resp.sendError(HttpServletResponse.SC_NOT_FOUND);
        
        String resolvedPath = path;
        
        if (this.aliases != null) {
            
            String alias = this.aliases.get(path);
            
            if (alias != null) {
             
                if (log.isDebugEnabled())
                    log.debug("Resolved ["+path+"] to alias ["+alias+"].");
                
                resolvedPath = alias;
            }
        }        
        
        try {
            
            String res = this.resourcePath + resolvedPath;
            
            if (log.isDebugEnabled())
                log.debug("Resolving resource ["+res+"]");
            
            is = this.getServletContext().getResourceAsStream(res);

            if (is == null) {
                
                log.error("Resource ["+res+"] not found.");
                resp.sendError(HttpServletResponse.SC_NOT_FOUND);
            }
            else {
                
                String inm = req.getHeader("If-None-Match");
                String[] clientETags = inm == null ? null : inm.split(",");
                
                if (log.isDebugEnabled())
                    log.debug("Client sent ETags ["+Arrays.toString(clientETags)+"]");
                
                String mimeType = this.getServletContext().getMimeType(resolvedPath);
                ResourceInfo ri = this.resourceInfos.get(resolvedPath);
          
                byte[] buf = null;
                
                if (ri == null) {
                    
                    if (log.isDebugEnabled())
                        log.debug("Calculating ResourceInfo for resource ["+resolvedPath+"]");

                    buf = new byte[this.bufferSize];
                    ri = this.enterETag(resolvedPath,is,buf);
                    is.close();
                    is = this.getServletContext().getResourceAsStream(res);
                }
                
                if (log.isDebugEnabled())
                    log.debug("Got ResourceInfo ["+ri+"]");
                
                boolean changed = true;
                
                if (clientETags != null) {
                    for (String clientETag : clientETags) {
                    
                        String ce = clientETag.trim();
                        
                        if (ce.equals(ri.eTag)) {
                            changed = false;
                            break;
                        }
                    }
                }
                
                if (mimeType != null)
                    resp.setContentType(mimeType);
                
                resp.setHeader("ETag",ri.eTag);
                
                if (changed) {
                    
                    resp.setHeader("Last-Modified",this.lastModified);
                    
                    if (buf == null)
                        buf = new byte[this.bufferSize];
                    
                    boolean compress = ri.contentLength > this.minCompressionSize &&
                    (this.excludeMimeTypes == null || !this.excludeMimeTypes.contains(mimeType));
                    
                    if (compress) {
                        os = CompressionHelper.setupResponseOutputStream(req,resp); 
                        
                        if (log.isDebugEnabled())
                            log.debug("compressed=["+CompressionHelper.isCompressed(os)+"].");
                    }
                    else {
                        
                        if (log.isDebugEnabled())
                            log.debug("Compression dissabled by configuration.");

                        os = resp.getOutputStream();
                    }
                    
                    if (!compress || !CompressionHelper.isCompressed(os))
                        // don't use resp.setContentLength(int), because it's restricted
                        // to 31 bits aka 2 GB of data.
                        resp.setHeader("Content-Length",Long.toString(ri.contentLength));
                    
                    int n;
                    
                    while ((n=is.read(buf)) > 0) {
                        
                        os.write(buf,0,n);
                    }

                    if (log.isDebugEnabled())
                        log.debug("Finished request ["+path+"]");
                }
                else {
                    if (log.isDebugEnabled())
                        log.debug("Sending "+HttpServletResponse.SC_NOT_MODIFIED+" not modified.");
       
                    resp.sendError(HttpServletResponse.SC_NOT_MODIFIED);
                }
            }
        }
        finally {
            if (os != null) CompressionHelper.finishOutputStream(os);
            if (is != null) is.close();
        }
    }

    /**
     * @return the internal buffer size.
     */
    public int getBufferSize()
    {
        return this.bufferSize;
    }

    /**
     * @param bufferSize the internal buffer size to set. The default value is 4096 and should be
     *                   suitable for almost any purpose.
     */
    public void setBufferSize(int bufferSize)
    {
        this.bufferSize = bufferSize;
    }

    /**
     * @return the resource path, which is exported from the given ClassLoader. If set to
     *           <code>null</code>, only redirections are available.
     */
    public String getResourcePath()
    {
        return this.resourcePath;
    }

    /**
     * @param resourcePath the resource path, which is exported from the given ClassLoader to set.
     *           All request exports resources acquired through
     *           {@link ServletContext#getResourceAsStream(String)} using this resource path
     *           appended by {@link HttpServletRequest#getPathInfo()}.
     */
    public void setResourcePath(String resourcePath)
    {
        this.resourcePath = resourcePath;
    }

    /**
     * @return the the minimal size of objects to be compressed.
     */
    public int getMinCompressionSize()
    {
        return this.minCompressionSize;
    }

    /**
     * @param minCompressionSize the minimal size of objects to be compressed to set.
     */
    public void setMinCompressionSize(int minCompressionSize)
    {
        this.minCompressionSize = minCompressionSize;
    }

    /**
     * @return The list of MIME types to be excluded from compression.
     */
    public Set<String> getExcludeMimeTypes()
    {
        return this.excludeMimeTypes;
    }

    /**
     * @param excludeMimeTypes The list of MIME types to be excluded from compression to set.
     *               Most image types are already compressed, so should consider to include
     *               their MIME types in this list.
     */
    public void setExcludeMimeTypes(Set<String> excludeMimeTypes)
    {
        this.excludeMimeTypes = excludeMimeTypes;
    }

    /**
     * Add an element to the list of MIME types to be excluded from compression.
     * @param excludeMimeType The MIME type to add.
     */
    public void addExcludeMimeType(String excludeMimeType) {
        
        if (this.excludeMimeTypes == null)
            this.excludeMimeTypes = new HashSet<String>();
        
        this.excludeMimeTypes.add(excludeMimeType);
    }
    
    /* (non-Javadoc)
     * @see javax.servlet.GenericServlet#getServletInfo()
     */
    @Override
    public String getServletInfo()
    {
        return "ResourceServlet[" + this.resourcePath +"]";
    }

    private static Map<String,String> parseMapping(String mapping) throws ServletException {
        
        String[] emt = mapping.split(",");
  
        Map<String,String> emts = new HashMap<String,String>();
        
        for (String mt : emt) {
            
            String[] x = mt.trim().split("=",2);
            
            if (x.length < 2)
                throw new ServletException("Invalid path mapping part ["+mt+"] specified (should contain an '=' character).");
                
            emts.put(x[0],x[1]);
        }
        return emts;
    }
    
    /* (non-Javadoc)
     * @see javax.servlet.GenericServlet#init(javax.servlet.ServletConfig)
     */
    @Override
    public void init() throws ServletException
    {
        ServletConfig cfg = this.getServletConfig();
        
        String p = cfg.getInitParameter("resourcePath");
        
        if (p != null)
            this.setResourcePath(p);
        
        p = cfg.getInitParameter("excludeMimeTypes");
        
        if (p != null) {
            
            String[] emt = p.split(",");
            Set<String> emts = new HashSet<String>();
            for (String mt : emt) {
                emts.add(mt.trim());
            }
            this.setExcludeMimeTypes(emts);
        }
        
        p = cfg.getInitParameter("bufferSize");
        
        if (p != null) {
            this.setBufferSize(Integer.parseInt(p));
        }
        
        p = cfg.getInitParameter("minCompressionSize");
        
        if (p != null) {
            this.setMinCompressionSize(Integer.parseInt(p));
        }
        
        p = cfg.getInitParameter("aliases");
        
        if (p != null) {
            this.setAliases(parseMapping(p));
        }
        
        p = cfg.getInitParameter("redirects");
        
        if (p != null) {
            this.setRedirects(parseMapping(p));
        }
    }

    /**
     * @return The mapping of requests path infos to redirect URLs.
     */
    public Map<String, String> getRedirects()
    {
        return this.redirects;
    }

    /**
     * @param redirects The mapping of requests path infos to redirect URLs to set.
     *         The keys of this map are path infos like returned by
     *         {@link HttpServletRequest#getPathInfo()}, which start with a '/' sign.
     *         The values are relative or absolute URLs, which are passed to
     *         {@link HttpServletResponse#sendRedirect(String)}.
     */
    public void setRedirects(Map<String, String> redirects)
    {
        this.redirects = redirects;
    }

    /**
     * Add a path-to-URL redirect.
     * 
     * @param pathInfo A path info as returned by {@link HttpServletRequest#getPathInfo()}.
     * @param url A redirect URL passed to {@link HttpServletResponse#sendRedirect(String)}.
     */
    public void addRedirect(String pathInfo, String url) {
        
        if (this.redirects == null)
            this.redirects= new HashMap<String, String>();
        
        this.redirects.put(pathInfo,url);
    }
    
    /**
     * @return the resource aliases.
     */
    public Map<String, String> getAliases()
    {
        return this.aliases;
    }

    /**
     * @param aliases the resource aliases to set.
     *         The keys of this map are path infos like returned by
     *         {@link HttpServletRequest#getPathInfo()}, which start with a '/' sign.
     *         The values are translated path infos and should therefore start with
     *         a '/' sign, too.
     */
    public void setAliases(Map<String, String> aliases)
    {
        this.aliases = aliases;
    }
    
    /**
     * Add an alias for a given path info.
     * 
     * @param pathInfo A path info as returned by {@link HttpServletRequest#getPathInfo()}.
     * @param alias An alias path info, which should start with a '/' sign.
     */
    public void addAlias(String pathInfo, String alias) {
        
        if (this.aliases == null)
            this.aliases = new HashMap<String, String>();
        
        this.aliases.put(pathInfo,alias);
    }
}
