/***********************************************************
 * $Id: BasicHttpContext.java 391 2010-10-08 16:52:49Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 18.09.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http.osgi;

import java.io.IOException;
import java.net.URL;
import java.util.Locale;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.osgi.framework.Bundle;
import org.osgi.service.http.HttpContext;

/**
 * A HttpContext implementation with an extension to MIME type mapping and resources
 * received from a specified bundle, which is usually injected using blueprint.   
 * 
 * <p>
 * The behavior of the <code>HttpContext</code> methods is
 * defined as follows:
 * <ul>
 * <li><code>getMimeType</code> Returns a MIME type based on the file extension of
 *  the given path, if an extensionMimeTypes map has been supplied. Otherwise
 * <code>null</code> is returned.
 * <li><code>handleSecurity</code> Always return <code>true</code>.
 * <li><code>getResource</code> Assumes the named resource is in the
 * context bundle; this method calls the context bundle's
 * <code>Bundle.getResource</code> method using the supplied resource prefix,
 * and returns the appropriate URL to
 * access the resource. On a Java runtime environment that supports
 * permissions, the Http Service needs to be granted 
 * <code>org.osgi.framework.AdminPermission[*,RESOURCE]</code>.
 * </ul></p>
 * 
 * <p>It is recommended to set up this bean inside a blueprint container like in the following
 * example:</p>
 * <pre>
 *  &lt;bean id="httpContext" class="org.clazzes.util.http.osgi.BasicHttpContext"&gt;
 *    &lt;property name="resourceBundle" ref="blueprintBundle"/&gt;
 *    &lt;property name="extensionMimeTypes"&gt;
 *      &lt;map&gt;
 *        &lt;entry key="html" value="text/html"/&gt;       
 *        &lt;entry key="xml"  value="text/xml"/&gt;       
 *        &lt;entry key="js"   value="application/javascript"/&gt;       
 *        &lt;entry key="json" value="application/json"/&gt;       
 *        &lt;entry key="css"  value="text/css"/&gt;       
 *        &lt;entry key="png"  value="image/png"/&gt;       
 *        &lt;entry key="jpg"  value="image/jpeg"/&gt;       
 *        &lt;entry key="jpeg" value="image/jpeg"/&gt;       
 *        &lt;entry key="gif"  value="image/gif"/&gt;       
 *        &lt;entry key="tif"  value="image/tiff"/&gt;       
 *      &lt;/map&gt;
 *    &lt;/property&gt;
 *  &lt;/bean&gt;
 * </pre>
 * 
 * <p>Please do not forget to manually import the <code>org.clazzes.util.http.osgi</code>
 * package in blueprint bundles, otherwise you will run into {@link ClassNotFoundException}s.
 * </p> 
 */
public class BasicHttpContext implements HttpContext
{
    private Bundle resourceBundle;
    private String resourcePrefix;
    private Map<String,String> extensionMimeTypes;
    
    /* (non-Javadoc)
     * @see org.osgi.service.http.HttpContext#handleSecurity(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    public boolean handleSecurity(HttpServletRequest request,
                                  HttpServletResponse response) throws IOException
    {
        return true;
    }

    /* (non-Javadoc)
     * @see org.osgi.service.http.HttpContext#getResource(java.lang.String)
     */
    public URL getResource(String name)
    {
        if (this.resourceBundle == null)
            return null;
        
        if (this.resourcePrefix == null)
            return this.resourceBundle.getResource(name);
        else
            return this.resourceBundle.getResource(this.resourcePrefix+name);            
    }

    /* (non-Javadoc)
     * @see org.osgi.service.http.HttpContext#getMimeType(java.lang.String)
     */
    public String getMimeType(String name)
    {
        String mimeType = null;
        
        if (this.extensionMimeTypes != null) {
        
            int dpos = name.lastIndexOf('.');

            if (dpos >= 0) {
         
                String ext = name.substring(dpos+1).toLowerCase(Locale.US);
                mimeType = this.extensionMimeTypes.get(ext);
            }
        }
        
        return mimeType;
    }

    /**
     * @return the bundle used to resolve resources. 
     */
    public Bundle getResourceBundle()
    {
        return this.resourceBundle;
    }

    /**
     * @param resourceBundle the bundle used to resolve resources to set.
     */
    public void setResourceBundle(Bundle resourceBundle)
    {
        this.resourceBundle = resourceBundle;
    }

    /**
     * @return A map containing file extension as keys and MIME types as values.
     */
    public Map<String, String> getExtensionMimeTypes()
    {
        return this.extensionMimeTypes;
    }

    /**
     * @param extensionMimeTypes A map containing file extension as keys and MIME types
     *           as values to set. The extensions are not case-sensitive, please supply
     *           supply the file extension in lower case. 
     */
    public void setExtensionMimeTypes(Map<String, String> extensionMimeTypes)
    {
        this.extensionMimeTypes = extensionMimeTypes;
    }

    /**
     * @return the prefix to attach when searching resource paths in {@link #getResource(String)}.
     */
    public String getResourcePrefix()
    {
        return this.resourcePrefix;
    }

    /**
     * @param resourcePrefix the prefix to attach when searching resource paths in
     *          {@link #getResource(String)} to set. If you set this property to
     *          <code>WEB-INF/</code>, all resources are searched inside the WEB-INF
     *          folder of your OSGi bundle, which is the recommended approach.
     */
    public void setResourcePrefix(String resourcePrefix)
    {
        this.resourcePrefix = resourcePrefix;
    }

}
